// cBuilding.cpp

#include "cBuilding.h"

#include "macros.h"

#include <GL/gl.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/glext.h>

#include <iostream>
using std::cout;

unsigned char facade_roof[] = {
    8, 9, 8, 9,
    7, 8, 7, 8,
    8, 9, 8, 9,
    7, 8, 7, 8
};

unsigned char facade[] = {
    6, 5, 5, 5, 5, 5, 5, 6,
    9, 2, 1, 1, 1, 1, 0, 9,
    9, 2, 2, 2, 1, 1, 1, 9,
    8, 7, 7, 7, 7, 7, 7, 8
};

unsigned char facade_damaged[] = {
    6, 5, 3, 5, 5, 3, 5, 6,
    8, 2, 1, 1, 4, 5, 0, 4,
    6, 2, 2, 4, 3, 1, 1, 8,
    4, 5, 5, 6, 5, 6, 4, 7
};

unsigned char facade_destroyed[] = {
    5, 4, 3, 5, 5, 3, 5, 6,
    6, 3, 1, 1, 4, 5, 3, 4,
    5, 2, 2, 4, 3, 1, 4, 6,
    3, 5, 5, 6, 5, 6, 4, 5
};

unsigned char concrete[] = {
    6, 6, 6, 6,
    6, 6, 6, 6,
    6, 6, 6, 6,
    6, 6, 6, 6
};

unsigned char junction[] = {
    6, 3, 3, 6,
    3, 3, 3, 3,
    3, 3, 3, 3,
    6, 3, 3, 6
};

unsigned char eastwest[] = {
    6, 6, 6, 6,
    3, 3, 3, 3,
    3, 3, 3, 3,
    6, 6, 6, 6
};

unsigned char northsouth[] = {
    6, 3, 3, 6,
    6, 3, 3, 6,
    6, 3, 3, 6,
    6, 3, 3, 6
};

static void textureGray2RGB(unsigned char* gray, unsigned char* rgb, int numpix) {
    unsigned char* p = rgb;

    loopi(numpix) {
        unsigned char g = (unsigned char) (gray[i] * 255.0f / 9.0f);
        *p++ = g;
        *p++ = g;
        *p++ = g;
    }
}

int cBuilding::sInstances = 0;
long cBuilding::sTextures[4];

cBuilding::cBuilding(int x, int y, int z, int rooms_x, int rooms_y, int rooms_z) {
    sInstances++;
    if (sInstances == 1) {
        unsigned int texname;
        const int WIDTH = 8;
        const int HEIGHT = 4;
        unsigned char texels[WIDTH * HEIGHT * 3];

        textureGray2RGB(facade, texels, WIDTH * HEIGHT);
        glBindTexture2D(texname, true, false, true, true, WIDTH, HEIGHT, texels);
        sTextures[0] = texname;

        textureGray2RGB(facade_damaged, texels, WIDTH * HEIGHT);
        glBindTexture2D(texname, true, false, true, true, WIDTH, HEIGHT, texels);
        sTextures[1] = texname;

        textureGray2RGB(facade_destroyed, texels, WIDTH * HEIGHT);
        glBindTexture2D(texname, true, false, true, true, WIDTH, HEIGHT, texels);
        sTextures[2] = texname;

        textureGray2RGB(facade_roof, texels, WIDTH / 2 * HEIGHT);
        glBindTexture2D(texname, true, false, true, true, WIDTH / 2, HEIGHT, texels);
        sTextures[3] = texname;
    }
    addRole(COLLIDEABLE);
    addRole(DAMAGEABLE);
    addRole(BUILDING);
    //vector_set(mPos, x * 3, y * 3, z * 3);
    vector_set(mRot, 0, 0, 0);
    quat_set(mOri, 0, 0, 0, 1);
    mRooms[0] = 3 * rooms_x;
    mRooms[1] = 3 * rooms_y;
    mRooms[2] = 3 * rooms_z;
    vector_set(mPos, x * 3 + mRooms[0]*0.5, y * 3, z * 3 + mRooms[2]*0.5);
    this->mRadius = sqrt(0.25 * mRooms[0] * mRooms[0] + 0.25 * mRooms[1] * mRooms[1] + 0.25 * mRooms[2] * mRooms[2]);
}

cObject::rImpact* cBuilding::damageByBeam(float* a, float* dir, float radius, float damage, bool all) {
    cObject::rImpact *result = NULL;
    float* p = this->mPos.data();
    float* w = this->mRooms;

    float dir_local[3];
    float a_local[3];
    {
        float ori_inv[4];
        quat_cpy(ori_inv, mOri);
        quat_conj(ori_inv);
        quat_apply(dir_local, ori_inv, dir);

        vector_cpy(a_local, a);
        vector_sub(a_local, a_local, p);
        quat_apply(a_local, ori_inv, a_local);
    }

    //float mins[3] = {0, 0, 0};
    //float maxs[3] = {+w[0], +w[1], +w[2]};
    float mins[3] = {-w[0]*0.5, 0, -w[2]*0.5};
    float maxs[3] = {+w[0]*0.5, +w[1], +w[2]*0.5};
    float impact[3];
    float normal[3];
    float delta = collideLineWithAABB(a_local, dir_local, mins, maxs, impact, normal);
    if (delta >= 0.0f) {
        //cout << "cBuilding::damageByBeam(): hit\n";
        if (damage > 0) {
            this->mHp[cObject::BODY] -= damage;
            if (this->mHp[cObject::BODY] <= 0 && !hasRole(DEAD)) {
                this->remRole(DAMAGEABLE);
                this->addRole(DEAD);
                cout << "cBuilding::damageByBeam(): DEAD\n";
                mExplosion.fire(0);
            }
        }

        result = new cObject::rImpact;
        {
            quat_apply(result->nrm, mOri, normal);

            quat_apply(result->pos, mOri, impact);
            vector_add(result->pos, result->pos, p);
        }
    }
    //cout << "delta: " << delta << "\n";
    //    vector_print(a);
    //    vector_print(a_local);
    //    vector_print(b);
    //    vector_print(b_local);
    return result;
}

void cBuilding::onSpawn() {
    mExplosion.mOwner = this;
}

void cBuilding::animate(float spf) {
    mExplosion.animate(spf);
}

void cBuilding::transform() {
    glPushMatrix();
    {
        float* p = this->mPos.data();
        float* r = this->mRot.data();
        float* w = this->mRooms;
        glTranslatef(p[0], p[1], p[2]);
        if (r[1] < -0.01f || r[1] > 0.01f) glRotatef(r[1], 0, 1, 0);
        //glTranslatef(0.5 * w[0], 0.5 * 0.5 * 0.5 * 3.5 * w[1], 0.5 * w[2]);
        glTranslatef(0, 0.5 * 0.5 * 0.5 * 3.5 * w[1], 0);
        glGetFloatv(GL_MODELVIEW_MATRIX, mBase);
        mExplosion.mBasefv = mBase;
    }
    glPopMatrix();
    mExplosion.transform();
}

void cBuilding::drawSolid() {
    mExplosion.drawSolid();

    float* p = this->mPos.data();
    float* r = this->mRot.data();
    float* w = this->mRooms;

    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    {
        glEnable(GL_CULL_FACE);
        glEnable(GL_TEXTURE_2D);
        long roof = sTextures[3];
        long wall = 0;
        if (this->mHp[cObject::BODY] <= 0) {
            glColor3f(0.6, 0.5, 0.4);
            wall = sTextures[2];
        } else if (this->mHp[cObject::BODY] <= 50) {
            glColor3f(0.6, 0.7, 0.8);
            wall = sTextures[1];
        } else {
            glColor3f(0.7, 0.8, 0.9);
            wall = sTextures[0];
        }
        glPushMatrix();
        {
            if (r[1] < -0.01f || r[1] > 0.01f) glRotatef(r[1], 0, 1, 0);
            glTranslatef(p[0]-w[0]*0.5, p[1], p[2]-w[2]*0.5);
            //glBlockFlat(w[0], w[1], w[2], wall, wall, wall, wall, roof, roof);
            const float n2 = 1.0f / sqrt(2);
            const float n3 = 1.0f / sqrt(3);
            glBindTexture(GL_TEXTURE_2D, roof);                   \
            glBegin(GL_QUADS);
            {
                glNormal3f(0, 1, 0);

                glTexCoord2f(0, 0);
                glVertex3f(0, w[1], 0);

                glTexCoord2f(0, w[2]);
                glVertex3f(0, w[1], w[2]);

                glTexCoord2f(w[0], w[2]);
                glVertex3f(w[0], w[1], w[2]);

                glTexCoord2f(w[0], 0);
                glVertex3f(w[0], w[1], 0);
            }
            glEnd();
            glBindTexture(GL_TEXTURE_2D, wall);                   \
            glBegin(GL_TRIANGLE_STRIP);
            {
                float tx = 0;

                glNormal3f(-n2, 0, -n2);
                glTexCoord2f(tx, 0);
                glVertex3f(0, 0, 0);

                glNormal3f(-n3, n3, -n3);
                glTexCoord2f(tx, w[1]);
                glVertex3f(0, w[1], 0);

                tx += w[0];

                glNormal3f(+n2, 0, -n2);
                glTexCoord2f(tx, 0);
                glVertex3f(w[0], 0, 0);

                glNormal3f(+n3, n3, -n3);
                glTexCoord2f(tx, w[1]);
                glVertex3f(w[0], w[1], 0);

                tx += w[2];

                glNormal3f(+n2, 0, +n2);
                glTexCoord2f(tx, 0);
                glVertex3f(w[0], 0, w[2]);

                glNormal3f(+n3, n3, +n3);
                glTexCoord2f(tx, w[1]);
                glVertex3f(w[0], w[1], w[2]);

                tx += w[0];

                glNormal3f(-n2, 0, +n2);
                glTexCoord2f(tx, 0);
                glVertex3f(0, 0, w[2]);

                glNormal3f(-n3, n3, +n3);
                glTexCoord2f(tx, w[1]);
                glVertex3f(0, w[1], w[2]);

                tx += w[2];

                glNormal3f(-n2, 0, -n2);
                glTexCoord2f(tx, 0);
                glVertex3f(0, 0, 0);

                glNormal3f(-n3, n3, -n3);
                glTexCoord2f(tx, w[1]);
                glVertex3f(0, w[1], 0);
            }
            glEnd();

            if (!true) glNormAxis();
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cBuilding::drawEffect() {
    mExplosion.drawEffect();
}


// -----------------------------------------------------------------

int cTile::sInstances = 0;
long cTile::sTextures[cTile::MAX_KINDS];
long cTile::sDisplaylists[cTile::MAX_KINDS * 2] = {0,};

cTile::cTile(int x, int y, int z, int kind) {
    sInstances++;
    if (sInstances == 1) {
        unsigned int texname;
        const int WIDTH = 4;
        const int HEIGHT = 4;
        unsigned char texels[WIDTH * HEIGHT * 3];

        textureGray2RGB(concrete, texels, WIDTH * HEIGHT);
        glBindTexture2D(texname, true, false, true, true, WIDTH, HEIGHT, texels);
        sTextures[KIND_CONCRETE] = texname;

        textureGray2RGB(eastwest, texels, WIDTH * HEIGHT);
        glBindTexture2D(texname, true, false, true, true, WIDTH, HEIGHT, texels);
        sTextures[KIND_ROAD_EASTWEST] = texname;

        textureGray2RGB(northsouth, texels, WIDTH * HEIGHT);
        glBindTexture2D(texname, true, false, true, true, WIDTH, HEIGHT, texels);
        sTextures[KIND_ROAD_NORTHSOUTH] = texname;

        textureGray2RGB(junction, texels, WIDTH * HEIGHT);
        glBindTexture2D(texname, true, false, true, true, WIDTH, HEIGHT, texels);
        sTextures[KIND_ROAD_JUNKTION] = texname;
    }
    if (0) if (sInstances == 1) {
            // First one initializes
            unsigned int texname;
            const int WIDTH = 8;
            const int HEIGHT = 8;
            const int DEPTH = 8;
            unsigned char texels[WIDTH * HEIGHT * DEPTH * 3];
            float s = 3;
            generateTexture3D(WIDTH, HEIGHT, DEPTH, texels, deltaFunc_NoisyPlasma, ColorFunc_Delta, ColorFunc_Delta, ColorFunc_Delta, 1 * s, 1 * s, 1 * s);
            //glBindTexture3D(texname, true, true, true, true, true, WIDTH, HEIGHT, DEPTH, texels);
            sTextures[0] = texname;
        }
    //    addRole(COLLIDEABLE); // !!
    mPos[0] = x * 9.0f + 4.5f;
    mPos[1] = y * 3.0f;
    mPos[2] = z * 9.0f + 4.5f;
    mRot[0] = mRot[1] = mRot[2] = 0.0f;
    mKind = kind;
}

cObject::rImpact* cTile::damageByBeam(float* a, float* dir, float radius, float damage, bool all) {
    cObject::rImpact *result = NULL;
    return result;
}

void cTile::drawSolid() {
    float* p = this->mPos.data();
    float* r = this->mRot.data();

    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    {
        long ground = sTextures[mKind];
        glColor3f(1, 1, 1);
        glPushMatrix();
        {
            if (r[1] < -0.01f || r[1] > 0.01f) glRotatef(r[1], 0, 1, 0);
            glTranslatef(p[0]-4.5f, p[1] + 0.08, p[2]-4.5f);
            GLfloat p[][3] = {
                {0, 0, 0},
                {9, 0, 0},
                {0, 0, 9},
                {9, 0, 9}
            };
            const float a = 0.01;
            const float b = 0.99;
            GLfloat t[][2] = {
                {a, a},
                {b, a},
                {b, b},
                {a, b},
            };
            glDisable(GL_CULL_FACE);
            glEnable(GL_TEXTURE_2D);
            glColor3f(1, 1, 1);
            glBindTexture(GL_TEXTURE_2D, ground);
            glBegin(GL_QUADS);
            glNormal3f(0, -1, 0);
            glVertex3fv2fv(p[0], t[1]);
            glVertex3fv2fv(p[1], t[0]);
            glVertex3fv2fv(p[3], t[3]);
            glVertex3fv2fv(p[2], t[2]);
            glEnd();
            if (!true) glNormAxis();
            glDisable(GL_LIGHTING);
            float h = 2;
            glColor3f(0.3, 0.3, 0.3);
            if (mKind == KIND_ROAD_JUNKTION && false) {
                glBegin(GL_LINES);
                glVertex3f(2, 0, 2);
                glVertex3f(2, h, 2);
                glVertex3f(7, 0, 2);
                glVertex3f(7, h, 2);
                glVertex3f(7, 0, 7);
                glVertex3f(7, h, 7);
                glVertex3f(2, 0, 7);
                glVertex3f(2, h, 7);
                glEnd();
            } else if (mKind == cTile::KIND_ROAD_EASTWEST) {
                glBegin(GL_LINE_STRIP);
                if ((int(mPos[0] / 9) & 1) == 0) {
                    glVertex3f(5, 0, 2);
                    glColor3f(0.7, 0.7, 0.6);
                    glVertex3f(5, h, 2);
                    glColor3f(1, 0.9, 0.7);
                    glVertex3f(5, h + 0.2 * h, 2 + 0.3 * h);
                } else {
                    glVertex3f(5, 0, 7);
                    glColor3f(0.7, 0.7, 0.6);
                    glVertex3f(5, h, 7);
                    glColor3f(1, 0.9, 0.7);
                    glVertex3f(5, h + 0.2 * h, 7 - 0.3 * h);
                }
                glEnd();
            } else if (mKind == cTile::KIND_ROAD_NORTHSOUTH) {
                glBegin(GL_LINE_STRIP);
                if ((int(mPos[2] / 9) & 1) == 0) {
                    glVertex3f(2, 0, 5);
                    glColor3f(0.7, 0.7, 0.6);
                    glVertex3f(2, h, 5);
                    glColor3f(1, 0.9, 0.7);
                    glVertex3f(2 + 0.3 * h, h + 0.2 * h, 5);
                } else {
                    glVertex3f(7, 0, 5);
                    glColor3f(0.7, 0.7, 0.6);
                    glVertex3f(7, h, 5);
                    glColor3f(1, 0.9, 0.7);
                    glVertex3f(7 - 0.3 * h, h + 0.2 * h, 5);
                }
                glEnd();
            }
        }
        glPopMatrix();
    }
    glPopAttrib();
}

cPadmap::cPadmap(float x, float z) {
    vector_set(mPos, x, 0, z);
    vector_set(mRot, 0, 0, 0);
    addRole(COLLIDEABLE);
    float h[16 * 16] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

        0, 0, 0, 1, 1, 4, 4, 3, 3, 2, 2, 2, 2, 0, 0, 0,
        0, 0, 0, 1, 1, 4, 4, 3, 3, 2, 2, 2, 2, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 3, 2, 2, 2, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 2, 2, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 3, 3, 0, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 3, 3, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 4, 4, 0, 0, 0, 0,

        0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 0, 0,
        0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    };

    memcpy(mHeights, h, sizeof (float) *16 * 16);
}

cObject::rImpact* cPadmap::damageByBeam(float* a, float* dir, float radius, float damage, bool all) {
    return NULL;
}

void cPadmap::drawSolid() {
    return; // !!

    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_TEXTURE_2D);
        //glDisable(GL_CULL_FACE);
        glColor4f(1, 1, 1, 1);

        int a = 16;
        int b = 16;

        /*
        glBegin(GL_TRIANGLE_STRIP);
        glVertex3f(0, 0, 0);
        glVertex3f(0, 0, b);
        glVertex3f(a, 0, 0);
        glVertex3f(a, 0, b);
        glEnd();
         */


        // Draw pad tops
        {
            glColor4f(0, 0, 1, 1);
            glNormal3f(0, 1, 0);
            int z = 0;
            for (int j = 0; j < b - 1; j += 2) {
                glBegin(GL_TRIANGLE_STRIP);
                int x = 0;
                for (int i = 0; i < a - 1; i += 2) {
                    bool stumble = (i == 0);
                    if (!stumble) {
                        stumble = (mHeights[(j + 0) * a + i - 1] != mHeights[(j + 0) * a + i])
                                || (mHeights[(j + 1) * a + i - 1] != mHeights[(j + 1) * a + i]);
                    }
                    if (stumble) {
                        glEnd();
                        glBegin(GL_TRIANGLE_STRIP);
                        glVertex3f(x, mHeights[(j + 0) * a + i], z);
                        glVertex3f(x, mHeights[(j + 1) * a + i], z + 1);
                    }
                    glVertex3f(x + 1, mHeights[(j + 0) * a + i + 1], z);
                    glVertex3f(x + 1, mHeights[(j + 1) * a + i + 1], z + 1);
                    x++;
                }
                glEnd();
                z++;
            }
        } // Draw pad tops

        // Draw east-west cliffs
        {
            glColor4f(1, 0, 1, 1);
            glNormal3f(0, 0, 1);
            int x = 0;
            for (int i = 0; i < a; i += 2) {
                int z = 1;
                for (int j = 2; j < b - 1; j += 2) {
                    bool stumble = (mHeights[(j - 1) * a + (i + 0)] != mHeights[(j + 0) * a + (i + 0)])
                            || (mHeights[(j - 1) * a + (i + 1)] != mHeights[(j + 0) * a + (i + 1)]);
                    if (stumble) {
                        glBegin(GL_TRIANGLE_STRIP);
                        glVertex3f(x + 1, mHeights[(j - 1) * a + i + 1], z);
                        glVertex3f(x + 0, mHeights[(j - 1) * a + i + 0], z);
                        glVertex3f(x + 1, mHeights[(j - 0) * a + i + 1], z);
                        glVertex3f(x + 0, mHeights[(j - 0) * a + i + 0], z);
                        glEnd();
                    }
                    z++;
                }
                x++;
            }
        } // Draw east-west cliffs

        // Draw north-south cliffs
        {
            glColor4f(0, 1, 1, 1);
            glNormal3f(1, 0, 0);
            int z = 0;
            for (int j = 0; j < b; j += 2) {
                int x = 1;
                for (int i = 2; i < a - 1; i += 2) {
                    bool stumble = (mHeights[(j + 0) * a + i - 1] != mHeights[(j + 0) * a + i])
                            || (mHeights[(j + 1) * a + i - 1] != mHeights[(j + 1) * a + i]);
                    if (stumble) {
                        glBegin(GL_TRIANGLE_STRIP);
                        glVertex3f(x, mHeights[(j + 0) * a + i - 1], z + 0);
                        glVertex3f(x, mHeights[(j + 1) * a + i - 1], z + 1);
                        glVertex3f(x, mHeights[(j + 0) * a + i + 0], z + 0);
                        glVertex3f(x, mHeights[(j + 1) * a + i + 0], z + 1);
                        glEnd();
                    }
                    x++;
                }
                z++;
            }
        } // Draw north-south cliffs

    }
    glPopAttrib();
}
