// cMech.cpp

#include "cObject.h"

#include "cMech.h"
#include "cController.h"
#include "cPad.h"
#include "macros.h"

#include <assert.h>
#include <GL/gl.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/glext.h>
#include <AL/al.h>
#include <AL/alut.h>

#include <vector>
using std::vector;

#include <iostream>
using std::cout;

DEFINE_glprintf


        int cMech::sInstances = 0;
long cMech::sTextures[1];

cMech::cMech(float* pos, float* rot) {
    sInstances++;
    if (sInstances == 1) {
        // First one initializes
        unsigned int texname = 0;
        const int WIDTH = 64;
        const int HEIGHT = 64;
        const int DEPTH = 64;
        unsigned char texels[WIDTH * HEIGHT * DEPTH * 3];
        //float s = 3;
        //generateTexture3D(WIDTH, HEIGHT, DEPTH, texels, deltaFunc_NoisyPlasma, ColorFunc_Delta,ColorFunc_Delta,ColorFunc_Delta, 1*s,0.2*s,1*s);
        //glBindTexture3D(texname, false,false, true,true,true, WIDTH,HEIGHT,DEPTH,texels);
        long t = 0;

        loopi(WIDTH) {

            loopj(HEIGHT) {

                loopk(DEPTH) {
                    int d = deltaFunc_NoisyPlasma(i, j, k);
                    int r = d, g = d, b = d;
                    if (d > 220) {
                        r += 140;
                        g += 120;
                        b += 100;
                    } else if (d > 200) {
                        r += 120;
                        g += 90;
                        b += 60;
                    }
                    texels[t++] = r;
                    texels[t++] = g;
                    texels[t++] = b;
                }
            }
        }
        //glBindTexture3D(texname, true, false, true, true, true, WIDTH, HEIGHT, DEPTH, texels);
        sTextures[0] = texname;
    }

    addRole(COLLIDEABLE);
    addRole(DAMAGEABLE);
    addRole(SOCIALISED);
    addRole(ENTITY);
    addRole(MECH);

    mController = new cController(this, true);

    if (pos != NULL) vector_cpy(mPos, pos);
    if (rot != NULL) vector_cpy(mRot, rot);
    vector_set(mTwr, 0, 0, 0);

    float axis[] = {0, 1, 0};
    quat_rotaxis(mOri, mRot[Y]*0.017454, axis);

    loopi(MAX_ROTATORS) {
        mRotators[i][0] = mRotators[i][1] = mRotators[i][2] = 0.0f;
    };

    mRadius = 1.8f;

    mCurrentWeapon = 0;
    mJetOn = 0;
    mSpeed = 0;
    mJumpstate = FALLING;

    mCameraState = 1;

    //
    ALuint buffer;
    //buffer = alutCreateBufferHelloWorld();
    buffer = alutCreateBufferFromFile("data/sound/pow.wav");
    alGenSources(1, &mSoundSource);
    alSourcei(mSoundSource, AL_BUFFER, buffer);
    alSourcef(mSoundSource, AL_PITCH, 1.0f);
    alSourcef(mSoundSource, AL_GAIN, 1.0f);
    alSourcefv(mSoundSource, AL_POSITION, mPos.data());
    alSourcefv(mSoundSource, AL_VELOCITY, mVel.data());
    alSourcei(mSoundSource, AL_LOOPING, AL_FALSE);

    //if (this->mSerial != 20280809030000001ULL) alSourcePlay(mSource);
    //alSourcePlay(mSoundSource);

    mComcom = new cComcom(this);
    mTarcom = new cTarcom(this);
    mSyscom = new cSyscom(this);
    mWepcom = new cWepcom(this);
    mForcom = new cForcom(this);
    mNavcom = new cNavcom(this);

    do_moveFor(NULL);
}

cMech::~cMech() {
    delete mComcom;
    delete mTarcom;
    delete mSyscom;
    delete mWepcom;
    delete mForcom;
    delete mNavcom;
    if (sInstances == 1) {
        // Last one cleans up

    }
    sInstances--;
}

void cMech::onSpawn() {
    this->mountWeapon((char*) "CTorsor", &mExplosion, false);
    if (hasRole(HUMANPLAYER)) alSourcePlay(mSoundSource);
    cout << "Mech spawned " << mSerial << "\n";
}

void cMech::multEyeMatrix() {
    float camera[16];
    float rot_inv[16];
    float pos_inv[16];

    glPushMatrix();
    {
        // Get Inverse Components of Head Matrix
        glLoadMatrixf(mHead);
        glGetTransposeInverseRotationMatrix(rot_inv);
        glGetInverseTranslationMatrix(pos_inv);
        // Compose Camera Matrix from inverse components
        glLoadIdentity();
        glMultMatrixf(rot_inv);
        glMultMatrixf(pos_inv);
        glGetFloatv(GL_MODELVIEW_MATRIX, camera);
        //matrix_print(camera);
    }
    glPopMatrix();

    if (abs(mCameraState) == 1) { // 1st
        glMultMatrixf(camera);
    } else if (abs(mCameraState) == 2) { // 3rd
        glTranslatef(0, 0, -5);
        glRotatef(15, 1, 0, 0);
        glMultMatrixf(camera);
    } else if (abs(mCameraState) == 3) { // 3rd Far
        glTranslatef(0, 0, -15);
        glRotatef(15, 1, 0, 0);
        glMultMatrixf(camera);
    } else if (abs(mCameraState) == 4) { // Reverse 3rd Far
        glTranslatef(0, 0, -15);
        glRotatef(15, 1, 0, 0);
        glRotatef(180, 0, 1, 0);
        glMultMatrixf(camera);
    } else if (abs(mCameraState) == 5) { // Map Near
        glTranslatef(0, 0, -50);
        glRotatef(90, 1, 0, 0);
        glMultMatrixf(pos_inv);
    } else if (abs(mCameraState) == 6) { // Map Far
        glTranslatef(0, 0, -100);
        glRotatef(90, 1, 0, 0);
        glMultMatrixf(pos_inv);
    } // if camerastate
}

void cMech::setAsAudioListener() {
    float s = -0.1;
    float vel[] = {
        mVel[0] * cWorld::instance->mSPF * s,
        mVel[1] * cWorld::instance->mSPF * s,
        mVel[2] * cWorld::instance->mSPF * s
    };
    float pos[] = {
        mPos[0], mPos[1], mPos[2]
    };
    alListenerfv(AL_POSITION, pos);
    alListenerfv(AL_VELOCITY, vel);
    float at_and_up[] = {
        mHead[8], mHead[9], mHead[10],
        0, -1, 0
    };
    alListenerfv(AL_ORIENTATION, at_and_up);
}

void cMech::ChassisLR(float r) {
    const float limit = 1.75;
    float e = r;
    if (e>+limit) e = +limit;
    if (e<-limit) e = -limit;
    mRot[Y] += e;
    mRot[Y] = fmod(mRot[Y], 360);
    if (mRot[Y] < 0) mRot[Y] += 360;

    float axis[] = {0, 1, 0};
    quat_rotaxis(mOri, mRot[Y]*0.017454, axis);
    //if (hasRole(HUMANPLAYER)) quat_print(mOri);
}

void cMech::ChassisUD(float r) {
    float e = r;
    if (e>+1) e = +1;
    if (e<-1) e = -1;
    mRot[X] += e;
    mRot[X] = fmod(mRot[X], 360);
    if (mRot[X] < 0) mRot[X] += 360;

    float axis[] = {1, 0, 0};
    quat_rotaxis(mOri, mRot[X]*0.017454, axis);
}

float cMech::TowerLR(float r) {
    float e = r;
    if (e>+2) e = +2;
    if (e<-2) e = -2;
    mTwr[Y] += e;
    if (this->hasRole(cObject::IMMOBILE)) {
        mTwr[Y] = fmod(mTwr[Y], 360);
        return 0.0f;
    } else {
        float excess = xsgn(mTwr[Y]) * (fabs(mTwr[Y]) > 75 ? fabs(mTwr[Y]) - 75 : 0);
        if (mTwr[Y] > +75) mTwr[Y] = +75;
        if (mTwr[Y] < -75) mTwr[Y] = -75;
        return excess;
    }
}

void cMech::TowerUD(float r) {
    float e = r;
    if (e>+2) e = +2;
    if (e<-2) e = -2;
    mTwr[X] += e;
    if (mTwr[X] > +75) mTwr[X] = +75;
    if (mTwr[X] < -75) mTwr[X] = -75;
}

void cMech::fireAllWeapons() {

    loopi(mWeapons.size()) {
        mWeapons[i]->fire(mTarget);
    }
}

void cMech::fireWeapon(unsigned n) {
    if (n >= mWeapons.size()) return;
    mWeapons[n]->fire(mTarget);
}

void cMech::mountWeapon(char* point, cWeapon *weapon, bool add) {
    if (strcmp(point, "LTorsor") == 0) weapon->mBasefv = mLeft;
    else if (strcmp(point, "LUpArm") == 0) weapon->mBasefv = mLeftupper;
    else if (strcmp(point, "LLoArm") == 0) weapon->mBasefv = mLeftlower;
    else if (strcmp(point, "RTorsor") == 0) weapon->mBasefv = mRight;
    else if (strcmp(point, "RUpArm") == 0) weapon->mBasefv = mRightupper;
    else if (strcmp(point, "RLoArm") == 0) weapon->mBasefv = mRightlower;
    else if (point[0] == 'L') weapon->mBasefv = mLeftupper;
    else if (point[0] == 'R') weapon->mBasefv = mRightupper;
    else weapon->mBasefv = mBack;

    weapon->mOwner = this;
    if (add) mWeapons.push_back(weapon);
}

void cMech::animatePhysics(float spf) {
    //const float e = 0.0174545;

    float height = checkRelHeight();
    bool near_ground = height < 0.3;

    bool jetting = mJetOn >= 15.0f;

    if (jetting && !near_ground) {
        mJumpstate = JUMPING;
    } else if (near_ground) {
        mJumpstate = GROUNDED;
    } else {
        mJumpstate = FALLING;
    }

    if (!true && hasRole(HUMANPLAYER)) {
        cout << "cMech::animateWalk(): Height over ground " << height << "\n";
        //cout << "Grounded " << mGrounded << "\n";
        cout << "JetOn: " << mJetOn << "\n";
    }

    vector_cpy(mRotators[CTorsor], mTwr);

    {
        const float mass = 11000;
        const float mass_inv = 1.0f / mass;

        // Reset acting forces.
        mAcc[X] = mAcc[Y] = mAcc[Z] = 0.0f;

        // Apply falling acceleration only acting when truly airborne.
        if (mJumpstate != GROUNDED) {

            loopi(3) {
                mAcc[i] = cWorld::instance->mGravity[i];
            }
        }

        // Jump Jet effect
        if (jetting || mJumpstate != GROUNDED) {
            const float jetforce = 140000.0f * (0.01f * mJetOn);
            mAcc[Y] += jetforce * mass_inv;
        }

        // Engine Effect
        //mAcc[X] += sin(mRot[Y] * e) * cos(mRot[X] * e) * driveforce;
        //mAcc[Z] += cos(mRot[Y] * e) * cos(mRot[X] * e) * driveforce;
        float driveforce = -170000.0 * mSpeed / 100.0f;
        if (mJumpstate != GROUNDED) driveforce *= 0.2;
        float fwd[] = {0, 0, driveforce * mass_inv};
        quat_apply(fwd, mOri, fwd);

        loopi(3) {
            mAcc[i] += fwd[i];
        }

        // Install a dampened spring when grounded.
        if (mJumpstate == GROUNDED) {
            mAcc[Y] += -height * 40.0;
            mAcc[Y] += -0.1f * mVel[Y] / spf;
        }

        // Prepare data for velocity dependent forces.
        float vel = vector_mag(mVel);
        float vel_norm[3];
        if (vel != 0.0f) {
            vector_scale(vel_norm, mVel, 1.0f / vel);
        } else {
            vel_norm[0] = vel_norm[1] = vel_norm[2] = 0.0f;
        }

#if 0
        // Ground friction
        if (mJumpstate == GROUNDED) {
            const float c = 3.3;
            const float f = c * mass * xmax(0, mAcc[Y]);
            float vel_norm[3];
            if (vel > 0.0001 && f < 0) {
                vector_scale(vel_norm, mVel, 1.0f / vel);

                loopi(3) {
                    mAcc[i] += vel_norm[i] * f * mass_inv;
                }
            }
        }
#else
        // Ground fricion through "damping".
        if (mJumpstate == GROUNDED) {
            const float friction = cWorld::instance->mGndfriction;

            loopi(3) {
                //mAcc[i] += -(1.0f - friction) * mVel[i];
                //mAcc[i] += -0.5 * mVel[i];
                //const float c = (1.0f - friction);//0.99999; //(1.0f - friction);
                //const float c_step = c * exp( -c * spf);
                const float c_step = friction / spf; //(1.0f - friction);
                mAcc[i] += -c_step * mVel[i];
            }
        }
#endif

        // Air drag
        {
            const float A = 1.5 * 3; // area [m2]
            const float p = cWorld::instance->mAirdensity; // ~1 [kg / m3]
            const float cw = 3*1.33; // not very aerodynamic

            float pstau = p * 0.5f * vel * vel;
            float F = cw * A * pstau;

            loopi(3) {
                mAcc[i] += F * -vel_norm[i] * mass_inv;
            }
        }


        // Integrate velocity.

        loopi(3) {
            mVel[i] += mAcc[i] * spf;
        }

        // Correct velocity when colliding so that
        // we directly arrive at possible impact point.
        checkAhead(spf);

        // Integrate position..

        loopi(3) {
            mPos[i] += mVel[i] * spf;
        }

    }
}

void cMech::poseRunning(float spf) {
    // Animate legs according to real velocity.
    float v[3];
    //vector_cpy(v, mVel);
    //vector_roty(v, -this->mRot[1]*0.017454f);
    float ori_inv[4];
    quat_cpy(ori_inv, mOri);
    quat_conj(ori_inv);
    quat_apply(v, ori_inv, mVel);

    float vel = sqrt(v[X] * v[X] + v[Z] * v[Z]) * -xsgn(v[Z]) * 0.16;
    if (vel > 2.8) vel = 2.8;
    if (vel < -1.0) vel = -1.0;

    // This is full of hand crafted magic numbers and magic code don't ask...
    const float e = 0.0174545;
    float o = e * -60;
    float t = e * mSeconds * 180;

    float l1 = 30 * (+sin(t)* 1) * vel;
    float l2 = 21 * (+sin(t * xsgn(vel) + o) + 1) * xabs(vel);

    float r1 = 30 * (-sin(t)* 1) * vel;
    float r2 = 21 * (-sin(t * xsgn(vel) + o) + 1) * xabs(vel);

    mRotators[LUpLeg][X] = -l1;
    mRotators[LLoLeg][X] = -l2;
    mRotators[RUpLeg][X] = -r1;
    mRotators[RLoLeg][X] = -r2;
}

void cMech::poseJumping(float spf) {
    // Hanging Legs
    const float a = -50;
    const float b = -130;
    const float s = 30;
    const float f = b / a;
    if (mRotators[LUpLeg][X] > -a) mRotators[LUpLeg][X] -= s * spf;
    if (mRotators[LUpLeg][X] < -a + 1) mRotators[LUpLeg][X] += s * spf;

    if (mRotators[LLoLeg][X] > +b) mRotators[LLoLeg][X] -= f * s * spf;
    if (mRotators[LLoLeg][X] < +b + 1) mRotators[LLoLeg][X] += f * s * spf;

    if (mRotators[RUpLeg][X] > -a) mRotators[RUpLeg][X] -= s * spf;
    if (mRotators[RUpLeg][X] < -a + 1) mRotators[RUpLeg][X] += s * spf;

    if (mRotators[RLoLeg][X] > +b) mRotators[RLoLeg][X] -= f * s * spf;
    if (mRotators[RLoLeg][X] < +b + 1) mRotators[RLoLeg][X] += f * s * spf;
}

void cMech::animate(float spf) {
    // Update time since spawn
    mSeconds += spf;

    if (mHp[cObject::BODY] <= 75) this->addRole(WOUNDED);
    if (mHp[cObject::BODY] <= 50) this->addRole(SERIOUS);
    if (mHp[cObject::BODY] <= 25) this->addRole(CRITICAL);
    //if (mHp[cObject::BODY] <= 0) this->addRole(DEAD);

    if (hasRole(DEAD)) {
        remRole(cObject::COLLIDEABLE);
        remRole(cObject::DAMAGEABLE);
        // Stop movement when dead
        mSpeed = 0;
        mJetOn = 0;
        // If already exploded then frag.
        // Animate Own Explosion (if fired)
        mExplosion.animate(spf);
        if (mExplosion.ready()) {
            this->ChassisLR(0);
            this->TowerLR(0);
            this->TowerUD(0);
            //cWorld::instance->fragObject(this);
        }
        if (mPad) mPad->reset();
    }

    // Process computers and Pad-input when not dead.
    if (!hasRole(DEAD)) {
        assert(mPad != NULL);
        //cout << mPad->toString();

        // Let all computers do their work.
        mComcom->process(spf);
        mTarcom->process(spf);
        mSyscom->process(spf);
        mWepcom->process(spf);
        mForcom->process(spf);
        mNavcom->process(spf);

        // Let AI-Controller set Pad-input.
        if (mController && mController->mEnabled) {
            if (mPad) mPad->reset();
            mController->process();
        }

        float excess = this->TowerLR(50 * spf * -mPad->getAxis(cPad::MECH_TURRET_LR_AXIS));
        this->TowerUD(50 * spf * mPad->getAxis(cPad::MECH_TURRET_UD_AXIS));

        if (!this->hasRole(cObject::IMMOBILE)) {
            // Steer left right
            this->ChassisLR(excess + 50 * spf * -mPad->getAxis(cPad::MECH_CHASSIS_LR_AXIS));
            // Speed
            float tdelta = 50 * spf * -mPad->getAxis(cPad::MECH_THROTTLE_AXIS);
            //cout << "tdelta: " << tdelta << "  speed: " << mSpeed << "\n";
            mSpeed = xmax(-100, tdelta*230);
            /*
            if (mSpeed > 0 && mSpeed + tdelta < 0) {
                mSpeed = mSpeed * 0.0;
            } else if (mSpeed < 0 && mSpeed + tdelta > 0) {
                mSpeed = mSpeed * 0.0;
            } else {
                mSpeed += 10 * tdelta;
                if (mSpeed > +220) mSpeed = +220;
                else
                    if (mSpeed < -100) mSpeed = -100;
            }
             */
        }

        if (mPad->getButton(cPad::MECH_FIRE_BUTTON)) {
            if (true) {
                mCurrentWeapon %= mWeapons.size();
                fireWeapon(mCurrentWeapon);
                mCurrentWeapon++;
                mCurrentWeapon %= mWeapons.size();
            } else {
                fireAllWeapons();
            }
        }

        if (mPad->getButton(cPad::MECH_JET_BUTTON)) {
            mJetOn += (100 - mJetOn) * 0.1;
        } else {
            mJetOn -= (mJetOn * 0.05);
        }

        if (mPad->getButton(cPad::MECH_CAMERA_BUTTON)) {
            // Only if Camera State is not transitional
            // then switch to next perspective on button press.
            if (mCameraState > 0) {
                mCameraState = (1 + (mCameraState % MAX_CAMERAMODES));
                mCameraState *= -1; // Set transitional.
            }
        } else {
            // Set Camera State as fixed.
            mCameraState = abs(mCameraState);
        }
    }

    // Rigid Body, Collisions etc.
    animatePhysics(spf);

    // Show fitting animation pose based on state and physics.
    if (mJumpstate == GROUNDED) {
        poseRunning(spf);
    } else {
        poseJumping(spf);
    }

    // Animate Weapons

    loopi(mWeapons.size()) {
        mWeapons[i]->animate(spf);
    }


    alSourcefv(mSoundSource, AL_POSITION, mPos.data());
    //alSourcefv(mSource, AL_VELOCITY, mVel);
}

/* ----------------------------------------------------------*/
// Model-Rendering-Code (should be evacuated sometime)

void drawLeft(bool transformOnly, float* upper, float* lower) {
    if (upper) {
        glPushMatrix();
        {
            glTranslatef(-0.25, 0, 0);
            glGetFloatv(GL_MODELVIEW_MATRIX, upper);
        }
        glPopMatrix();
    }
    if (lower) {
        glPushMatrix();
        {
            glTranslatef(-0.38, -0.6, -0.0);
            glGetFloatv(GL_MODELVIEW_MATRIX, lower);
        }
        glPopMatrix();
    }
    if (transformOnly) return;

    //glAxis(0.1);

    //float a[] = { 0.3, 0.3, 0.1 };
    //float b[] = { 0.1, 0.1, 0.1 };
    float a[] = {0.9, 0.9, 0.9};
    float b[] = {0.7, 0.7, 0.7};

    glPushMatrix();
    { // Shoulder Model
        glTranslatef(-0.2, 0.0, 0.0);
        glScalef(0.2, 0.1, 0.1);
        glColor3fv(b);
        glCenterUnitBlock();
    }
    glPopMatrix();

    glTranslatef(-0.35, 0, 0);
    //glAxis(0.1);

    glPushMatrix();
    { // Upper Arm Model
        glTranslatef(-0.05, -0.3, -0.15);
        glScalef(0.1, 0.3, 0.3);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();

    glTranslatef(+0.0, -0.4, 0);
    //glAxis(0.1);
    glRotatef(90, 1, 0, 0);
    glTranslatef(+0.0, -0.52, 0);
    //glAxis(0.1);
}

void drawRight(bool transformOnly, float* upper, float* lower) {
    if (upper) {
        glPushMatrix();
        {
            glTranslatef(+0.25, 0, 0);
            glGetFloatv(GL_MODELVIEW_MATRIX, upper);
        }
        glPopMatrix();
    }
    if (lower) {
        glPushMatrix();
        {
            glTranslatef(+0.38, -0.6, -0.0);
            glGetFloatv(GL_MODELVIEW_MATRIX, lower);
        }
        glPopMatrix();
    }
    if (transformOnly) return;

    //glAxis(0.1);

    //float a[] = { 0.3, 0.3, 0.1 };
    //float b[] = { 0.1, 0.1, 0.1 };
    float a[] = {0.9, 0.9, 0.9};
    float b[] = {0.7, 0.7, 0.7};

    glPushMatrix();
    { // Shoulder Model
        glTranslatef(+0.2, 0.0, 0.0);
        glScalef(0.2, 0.1, 0.1);
        glColor3fv(b);
        glCenterUnitBlock();
    }
    glPopMatrix();

    glTranslatef(+0.35, 0, 0);
    //glAxis(0.1);

    glPushMatrix();
    { // Upper Arm Model
        glTranslatef(+0.05, -0.3, -0.15);
        glScalef(0.1, 0.3, 0.3);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();

    glRotatef(0, 1, 0, 0);
    glTranslatef(+0.0, -0.4, 0);
    //glAxis(0.1);
    // Lower Arm

    glRotatef(90, 1, 0, 0);
    glTranslatef(+0.0, -0.52, 0);
    //glAxis(0.1);
    // Hand
}

void drawCenter(bool transformOnly, float xrot, float yrot, float* back, float* left, float* right, float* head) {
    glRotatef(yrot, 0, 1, 0);
    glRotatef(xrot, 1, 0, 0);
    if (head) {
        glPushMatrix();
        {
            glTranslatef(+0.0, +0.97, +0.0);
            glGetFloatv(GL_MODELVIEW_MATRIX, head);
        }
        glPopMatrix();
    }
    if (back) {
        glPushMatrix();
        {
            glTranslatef(+0.0, +0.6, +0.3);
            glGetFloatv(GL_MODELVIEW_MATRIX, back);
        }
        glPopMatrix();
    }
    if (left) {
        glPushMatrix();
        {
            glTranslatef(-0.7, +0.6, +0.2);
            glGetFloatv(GL_MODELVIEW_MATRIX, left);
        }
        glPopMatrix();
    }
    if (right) {
        glPushMatrix();
        {
            glTranslatef(+0.7, +0.6, +0.2);
            glGetFloatv(GL_MODELVIEW_MATRIX, right);
        }
        glPopMatrix();
    }
    if (transformOnly) return;

    //glAxis(0.1);

    //float a[] = { 0.3, 0.3, 0.1 };
    //float b[] = { 0.1, 0.1, 0.1 };
    float a[] = {0.9, 0.9, 0.9};
    float b[] = {0.7, 0.7, 0.7};

    // Center
    glPushMatrix();
    {
        glScalef(0.2, 0.39, 0.4);
        glTranslatef(0, 1, -0.5);
        glColor3fv(b);
        glCenterUnitBlock();
    }
    glPopMatrix();
    // Head
    glPushMatrix();
    {
        glTranslatef(0, 0.94, -0.1);
        glScalef(0.2, 0.16, 0.2);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();
    // Left Center
    glPushMatrix();
    {
        glTranslatef(-0.39, 0.58, 0.0);
        glScalef(0.3, 0.25, 0.35);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();
    // Right Center
    glPushMatrix();
    {
        glTranslatef(+0.39, 0.58, 0.0);
        glScalef(0.3, 0.25, 0.35);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();
}

void drawLower(bool transformOnly, float lhp, float lne, float rhp, float rne, float* center) {
    if (center) {
        glGetFloatv(GL_MODELVIEW_MATRIX, center);
    }
    if (transformOnly) return;

    //glAxis(0.1);

    //float a[] = { 0.3, 0.3, 0.1 };
    //float b[] = { 0.1, 0.1, 0.1 };
    float a[] = {0.9, 0.9, 0.9};
    float b[] = {0.7, 0.7, 0.7};

    glPushMatrix();
    { // Hips Model
        glTranslatef(+0.0, 0.0, 0.0);
        glScalef(0.3, 0.12, 0.12);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();

    // Left Leg
    glPushMatrix();
    {
        glTranslatef(-0.32, 0, 0);
        glRotatef(lhp, 1, 0, 0);
        //glAxis(0.1);
        glPushMatrix();
        { // Upper Leg Model
            glTranslatef(-0.0, -0.4, 0.0);
            glScalef(0.18, 0.5, 0.2);
            glColor3fv(b);
            glCenterUnitBlock();
        }
        glPopMatrix();
        glTranslatef(-0, -0.9, 0);
        // Lower Leg
        glPushMatrix();
        {
            glRotatef(lne, 1, 0, 0);
            //glAxis(0.1);
            glPushMatrix();
            { // Lower Leg Model
                glTranslatef(-0.0, -0.3, 0.0);
                glScalef(0.1, 0.4, 0.1);
                glColor3fv(a);
                glCenterUnitBlock();
            }
            glPopMatrix();
            glTranslatef(-0, -0.9, 0);
            //glAxis(0.1);
            glPushMatrix();
            { // Foot Model
                glTranslatef(0.0, +0.1, -0.09);
                glScalef(0.18, 0.1, 0.26);
                glColor3fv(b);
                glCenterUnitBlock();
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
    glPopMatrix();

    // Right Leg
    glPushMatrix();
    {
        glTranslatef(+0.32, 0, 0);
        glRotatef(rhp, 1, 0, 0);
        //glAxis(0.1);
        glPushMatrix();
        { // Upper Leg Model
            glTranslatef(+0.0, -0.4, 0.0);
            glScalef(0.18, 0.5, 0.2);
            glColor3fv(b);
            glCenterUnitBlock();
        }
        glPopMatrix();
        glTranslatef(+0, -0.9, 0);
        // Lower Leg
        glPushMatrix();
        {
            glRotatef(rne, 1, 0, 0);
            //glAxis(0.1);
            glPushMatrix();
            { // Lower Leg Model
                glTranslatef(+0.0, -0.3, 0.0);
                glScalef(0.1, 0.4, 0.1);
                glColor3fv(a);
                glCenterUnitBlock();
            }
            glPopMatrix();
            glTranslatef(+0, -0.9, 0);
            //glAxis(0.1);
            glPushMatrix();
            { // Foot Model
                glTranslatef(0.0, +0.1, -0.09);
                glScalef(0.18, 0.1, 0.26);
                glColor3fv(b);
                glCenterUnitBlock();
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
    glPopMatrix();
}

/* ----------------------------------------------------------*/


void cMech::transform() {
    if (true) {
        glPushMatrix();
        {
            glTranslatef(this->mPos[0], this->mPos[1] + 1.8, this->mPos[2]);
            //glRotatef(this->mRot[1], 0, 1, 0);
            glRotateq(this->mOri.data());
            if (hasRole(DEAD)) {
                glTranslatef(0, -1.6, 0);
                glRotatef(-40, 4, 1, 3);
            }

            bool transformOnly = true;

            float twx = this->mTwr[0];
            float twy = this->mTwr[1];
            float lhp = this->mRotators[LUpLeg][X];
            float lne = this->mRotators[LLoLeg][X];
            float rhp = this->mRotators[RUpLeg][X];
            float rne = this->mRotators[RLoLeg][X];

            glPushMatrix();
            {
                drawLower(transformOnly, lhp, lne, rhp, rne, mCenter);
                glLoadMatrixf(mCenter);
                drawCenter(transformOnly, twx, twy, mBack, mLeft, mRight, mHead);
                glLoadMatrixf(mLeft);
                drawLeft(transformOnly, mLeftupper, mLeftlower);
                glLoadMatrixf(mRight);
                drawRight(transformOnly, mRightupper, mRightlower);
                glLoadMatrixf(mBack);
            }
            glPopMatrix();

        }
        glPopMatrix();
    }

    loopi(mWeapons.size()) {
        mWeapons[i]->transform();
    }
}

void cMech::drawSolid() {
    // Drawing of model should start at the hips as base (0,0,0).
    if (true) {
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        {
            if (ENABLE_TEXTURE_3D) glEnable(GL_TEXTURE_3D);
            glBindTexture(GL_TEXTURE_3D, sTextures[0]);
            glMatrixMode(GL_TEXTURE);
            glMatrixMode(GL_MODELVIEW);

            glPushMatrix();
            {
                glTranslatef(this->mPos[0], this->mPos[1] + 1.8, this->mPos[2]);
                //glRotatef(this->mRot[1], 0, 1, 0);
                glRotateq(this->mOri.data());
                if (hasRole(DEAD)) {
                    glTranslatef(0, -1.6, 0);
                    glRotatef(-40, 4, 1, 3);
                }

                // Local-To-World Matrices (Bases) of Mountpoints.
                float head[16];
                float center[16];
                float back[16];
                float left[16];
                float right[16];
                float rightupper[16];
                float rightlower[16];
                float leftupper[16];
                float leftlower[16];

                bool transformOnly = false;

                float twx = this->mTwr[0];
                float twy = this->mTwr[1];
                float lhp = this->mRotators[LUpLeg][X];
                float lne = this->mRotators[LLoLeg][X];
                float rhp = this->mRotators[RUpLeg][X];
                float rne = this->mRotators[RLoLeg][X];

                //glAxis(0.15);
                glPushMatrix();
                {
                    drawLower(transformOnly, lhp, lne, rhp, rne, center);
                    glLoadMatrixf(center);
                    drawCenter(transformOnly, twx, twy, back, left, right, head);
                    glLoadMatrixf(left);
                    drawLeft(transformOnly, leftupper, leftlower);
                    glLoadMatrixf(right);
                    drawRight(transformOnly, rightupper, rightlower);
                    glLoadMatrixf(back);
                }
                glPopMatrix();

            }
            glPopMatrix();
        }
        glPopAttrib();
    }

    loopi(mWeapons.size()) {
        mWeapons[i]->drawSolid();
    }

    mExplosion.drawSolid();
}

void cMech::drawEffect() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    if (!mRoles[HUMANPLAYER]) {
        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glDepthMask(GL_FALSE);
        //glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        if (mRoles[RED]) glColor4f(1, 0, 0, 0.4);
        else if (mRoles[GREEN]) glColor4f(0, 1, 0, 0.4);
        else if (mRoles[BLUE]) glColor4f(0, 0, 1, 0.4);
        else glColor4f(1, 1, 0, 0.4);
        glPushMatrix();
        {
            glMultMatrixf(this->mCenter);
            glTranslatef(0, -1.7, 0);
            glRotatef(90, 1, 0, 0);
            glDisk(7, 1.3f);
        }
        glPopMatrix();
        glPushMatrix();
        {
            glMultMatrixf(this->mHead);
            glTranslatef(0, +0.9, 0);
            float s = 0.65;
            glScalef(s, s, s);
            float n[16];
            glGetTransposeInverseRotationMatrix(n);
            glMultMatrixf(n);
            int l = mName.length();
            glTranslatef(-l * 0.5f, 0, 0);
            glprintf(mName.c_str());
        }
        glPopMatrix();
    }
    glPopAttrib();

    // Draw jumpjet exaust if jet is somewhat on.
    if (this->mJetOn > 30.0f) {
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        {
            glDisable(GL_CULL_FACE);
            glDisable(GL_LIGHTING);
            glDisable(GL_TEXTURE_2D);
            glDepthMask(GL_FALSE);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE);

            glPushMatrix();
            {
                float f = mJetOn * 0.003;
                glMultMatrixf(this->mCenter);
                glTranslatef(0.3, -0.18, +0.28);
                float n[16];
                glGetTransposeInverseRotationMatrix(n);
                glMultMatrixf(n);
                glColor4f(1, 1, 0.3, 0.6);
                glDisk(7, f + 0.0003 * (rand() % 100));
                glColor4f(1, 0.5, 0.3, 0.7);
                glDisk(7, f * 0.6 + 0.001 * (rand() % 100));
                glColor4f(1, 1, 1, 0.8);
                glDisk(7, f * 0.3 + 0.001 * (rand() % 100));
            }
            glPopMatrix();
            glPushMatrix();
            {
                float f = mJetOn * 0.003;
                glMultMatrixf(this->mCenter);
                glTranslatef(-0.3, -0.18, +0.28);
                float n[16];
                glGetTransposeInverseRotationMatrix(n);
                glMultMatrixf(n);
                glColor4f(1, 1, 0.3, 0.6);
                glDisk(7, f + 0.0003 * (rand() % 100));
                glColor4f(1, 0.5, 0.3, 0.7);
                glDisk(7, f * 0.6 + 0.001 * (rand() % 100));
                glColor4f(1, 1, 1, 0.8);
                glDisk(7, f * 0.3 + 0.001 * (rand() % 100));
            }
            glPopMatrix();
            glPushMatrix();
            {
                float f = mJetOn * 0.003;
                glMultMatrixf(this->mBack);
                glTranslatef(-0.6, -0.2, +0.1);
                float n[16];
                glGetTransposeInverseRotationMatrix(n);
                glMultMatrixf(n);
                glColor4f(1, 1, 0.3, 0.6);
                glDisk(7, f + 0.0003 * (rand() % 100));
                glColor4f(1, 0.5, 0.3, 0.7);
                glDisk(7, f * 0.6 + 0.001 * (rand() % 100));
                glColor4f(1, 1, 1, 0.8);
                glDisk(7, f * 0.3 + 0.001 * (rand() % 100));
            }
            glPopMatrix();
            glPushMatrix();
            {
                float f = mJetOn * 0.003;
                glMultMatrixf(this->mBack);
                glTranslatef(+0.6, -0.2, +0.1);
                float n[16];
                glGetTransposeInverseRotationMatrix(n);
                glMultMatrixf(n);
                glColor4f(1, 1, 0.3, 0.6);
                glDisk(7, f + 0.0003 * (rand() % 100));
                glColor4f(1, 0.5, 0.3, 0.7);
                glDisk(7, f * 0.6 + 0.001 * (rand() % 100));
                glColor4f(1, 1, 1, 0.8);
                glDisk(7, f * 0.3 + 0.001 * (rand() % 100));
            }
            glPopMatrix();
        }
        glPopAttrib();
    }
    // Draw weapon's effects.

    loopi(mWeapons.size()) {
        mWeapons[i]->drawEffect();
    }
    // Draw effect of own explosion (yes, it is only when exploding).
    mExplosion.drawEffect();
}

void cMech::drawHUD() {
    glPushAttrib(GL_ALL_ATTRIB_BITS /* more secure */);
    {
        glLineWidth(2);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_FOG);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_CULL_FACE);

        glPushUnitOrthoProjection();
        {
            glPushMatrix();
            {
                glLoadIdentity();

                float bk[] = {0.0, 0.2, 0.3, 0.2};
                //float bk[] = {0, 0, 0, 0.0};
                float w = 5;
                float h = 4;
                float sx = 1.0f / w;
                float sy = 1.0f / h;

                cComputer * displays[4][5] = {
                    { mNavcom, NULL, NULL, NULL, mComcom},
                    { NULL, NULL, NULL, NULL, NULL},
                    { NULL, NULL, NULL, NULL, NULL},
                    { mTarcom, NULL, mWepcom, NULL, mSyscom}
                };

                loopj(4) {

                    loopi(5) {
                        if (displays[j][i] == NULL) continue;
                        glPushMatrix();
                        {
                            glScalef(sx, sy, 1);
                            glTranslatef(i, 3 - j, 0);
                            glColor4fv(bk);
                            displays[j][i]->drawHUD();
                        }
                        glPopMatrix();
                    }
                }

                if (true) {
                    glPushMatrix();
                    {
                        glTranslatef(0.25, 0.25, 0);
                        glScalef(0.5, 0.5, 1);
                        glTranslatef(0, 0, 0);
                        glColor4fv(bk);
                        mForcom->drawHUD();
                    }
                    glPopMatrix();
                }

            }
            glPopMatrix();
        }
        glPopProjection();
    }
    glPopAttrib();
}

cObject::rImpact* cMech::damageByBeam(float* a, float* dir, float radius, float damage, bool all) {
    cObject::rImpact *result = NULL;
    float* p = mPos.data();

    float dir_local[3];
    float a_local[3];
    {
        float ori_inv[4];
        quat_cpy(ori_inv, mOri);
        quat_conj(ori_inv);
        quat_apply(dir_local, ori_inv, dir);

        vector_cpy(a_local, a);
        vector_sub(a_local, a_local, p);
        quat_apply(a_local, ori_inv, a_local);
    }

    float impact[3];
    float normal[3];
    float delta = -1;
    int hitzone = cObject::BODY;

    {
        float ext_radius = 1.0f + radius;
        float height = 3.0f;
        float base[3] = {0, 0, 0};
        //delta = collideLineWithBillboard(a_local, dir_local, base, ext_radius, height, impact, normal);
        delta = collideLineWithCylinder(a_local, dir_local, base, ext_radius, height, impact, normal);
        if (impact[1] < 1.3) hitzone = cObject::LEGS;
        else if (impact[0] < -0.5 && this->mHp[cObject::LEFT] > 0) hitzone = cObject::LEFT;
        else if (impact[0] > +0.5 && this->mHp[cObject::RIGHT] > 0) hitzone = cObject::RIGHT;
    }

    if (delta >= 0.0f) {
        //cout << "cMech::damageByBeam(): hit\n";
        if (damage != 0.0f) {
            this->mHp[hitzone] -= damage;
            if (this->mHp[hitzone] < 0) this->mHp[hitzone] = 0;
            if (this->mHp[cObject::BODY] <= 0 && !this->hasRole(cObject::DEAD)) {
                this->remRole(DAMAGEABLE);
                this->remRole(COLLIDEABLE);
                this->addRole(DEAD);
                cout << "cMech::damageByBeam(): DEAD\n";
                mExplosion.fire(0);
            }
        }

        result = new cObject::rImpact;
        {
            quat_apply(result->nrm, mOri, normal);

            quat_apply(result->pos, mOri, impact);
            vector_add(result->pos, result->pos, p);
        }
    }
    return result;
}

tSerial cMech::enemyNearby() {
    tSerial result = 0;
    // Filter objects by distance
    std::list<cObject*>* scan = cWorld::instance->filterByRange(this, this->mPos.data(), 0.0f, 50.0f, -1, NULL);
    if (!scan->empty()) {
        // Find all objects belonging to any enemy party/role.
        std::list<cObject*>* roles = cWorld::instance->filterByRole(this, this->mEnemies.to_ulong(), false, -1, scan);
        if (!roles->empty()) {
            // Only those objects which are damageable.
            std::list<cObject*>* damages = cWorld::instance->filterByRole(this, (1UL << cObject::DAMAGEABLE), false, 1, roles);
            //cout << cWorld::instance->getNames(damages) << endl;
            if (!damages->empty()) result = damages->front()->mSerial;
            delete damages;
        }
        delete roles;
    }
    delete scan;
    return result;
}

float cMech::inDestinationRange() {
    float a = 0;
    float b = 6;
    float d = vector_distance(mPos, mDestination);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}

float cMech::inMeeleRange() {
    cObject* target = cWorld::instance->mIndex[mTarget];
    if (target == NULL) return 0.0f;
    float a = 16;
    float b = 24;
    float d = vector_distance(mPos, target->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}

float cMech::inWeaponRange() {
    // TODO should depend on ready to fire weapons properties.
    cObject* target = cWorld::instance->mIndex[mTarget];
    if (target == NULL) return 0.0f;
    float a = 36 + 10;
    float b = 44 + 10;
    float d = vector_distance(mPos, target->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}

float cMech::inTargetRange() {
    cObject* target = cWorld::instance->mIndex[mTarget];
    if (target == NULL) return 0.0f;
    float a = 56;
    float b = 64;
    float d = vector_distance(mPos, target->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}

// Calculate nearest signed rotation angle towards target position.
// First and second axis should be X (0) and Z (2) for angle around Y (1).
#define rotationTo(own_pos, target_pos, own_angle, first_axis, second_axis) \
({ \
    float r = 0.0f; \
    if (target_pos != NULL) { \
        const float e = 180.0f/3.145479f; \
        float dir[3] = { \
            (own_pos)[0] - (target_pos)[0], \
            (own_pos)[1] - (target_pos)[1], \
            (own_pos)[2] - (target_pos)[2]  \
        }; \
        float tgt_angle = atan2(dir[first_axis], dir[second_axis]) * e; \
        float delta = tgt_angle - (own_angle); \
        r = delta; \
        if (fabs(delta) > 180) r = -(delta / fabs(delta)) * (360 - fabs(delta)); \
        if (!true) cout << "own: " << own_angle << "  tgt: " << tgt_angle << "  delta: " << delta << "  result: " << r << "\n"; \
        r *= -1.0f; \
    } \
    r; \
})

void rotationTo_(float* result2f, float* own_pos, float* tgt_pos, float* base_ori, float*tower_ori = NULL) {
    float dir_global[3];
    vector_sub(dir_global, tgt_pos, own_pos);
    float ori_inv[4];
    quat_cpy(ori_inv, base_ori);
    if (tower_ori) quat_mul(ori_inv, ori_inv, tower_ori);
    quat_conj(ori_inv);
    float dir_local[4];
    quat_apply(dir_local, ori_inv, dir_global);
    vector_norm(dir_local, dir_local);
    result2f[X] = dir_local[X];
    if (dir_local[Z] > 0) result2f[X] += xsgn(result2f[X]);
    result2f[Y] = dir_local[Y];
}

void cMech::do_moveTowards() {
    //cout << "do_moveTowards():\n";
    if (mPad == NULL) return;

    // Determine Target (ie. Move- or Aim-Target if former is not available).
    float* target_pos = NULL;
    if (finitef(mDestination[0])) {
        target_pos = mDestination.data();
    } else if (mTarget != 0) {
        cObject* target = cWorld::instance->mIndex[mTarget];
        if (target != NULL) {
            target_pos = target->mPos.data();
        } else {
            mTarget = 0;
            return;
        }
    } else return;

    // Determine nearest rotation direction
    //float rd = rotationTo(mPos, target_pos, this->mRot[Y], X, Z);
    float rd[2];
    rotationTo_(rd, mPos.data(), target_pos, this->mOri.data());
    mPad->setAxis(cPad::MECH_CHASSIS_LR_AXIS, rd[X]);

    //float thr = 20.0f * (1.0f - fabs(rd[X]) / 360.0f);
    float thr = 1.0f * (1.0f - 0.6*fabs(rd[X]));
    //cout << "Throttle: " << thr << "\n";
    //mPad->setAxis(cPad::MECH_THROTTLE_AXIS, (char) - thr);
    mPad->setAxis(cPad::MECH_THROTTLE_AXIS, -thr);
}

void cMech::do_moveNear() {
    //cout << "do_moveNear():\n";
    if (mPad == NULL) return;

    // Determine Target (ie. Move- or Aim-Target if former is not available).
    float* target_pos = NULL;
    if (finitef(mDestination[0])) {
        target_pos = mDestination.data();
    } else if (mTarget != 0) {
        cObject* target = cWorld::instance->mIndex[mTarget];
        if (target != NULL) {
            target_pos = target->mPos.data();
        } else {
            mTarget = 0;
            return;
        }
    } else return;

    // Determine nearest rotation direction
    //float rd = rotationTo(mPos, target_pos, this->mRot[Y], X, Z);
    float rd[2];
    rotationTo_(rd, mPos.data(), target_pos, this->mOri.data());
    mPad->setAxis(cPad::MECH_CHASSIS_LR_AXIS, 2*rd[X]);

    float d = vector_distance(this->mPos, target_pos);

    float thr = 1.0f * (1.0f - 0.7*fabs(rd[X]));

    float f = (d - 23);
    if (f < -1) f = -1;
    if (f > +1) f = +1;

    f *= thr;

    mPad->setAxis(cPad::MECH_THROTTLE_AXIS, -f);
    //cout << "Throttle: " << thr << "\n";
    //mPad->setAxis(cPad::MECH_THROTTLE_AXIS, (char) - thr);
}

void cMech::do_aimAt() {
    //cout "do_aimAt():\n";
    if (mPad == NULL) return;

    // Get aim-target position.
    float* target_pos = NULL;
    if (mTarget != 0) {
        cObject* target = cWorld::instance->mIndex[mTarget];
        if (target != NULL) target_pos = target->mPos.data();
        else return;
    } else return;

    // Determine nearest rotation direction
    //float rd = rotationTo(mPos, target_pos, this->mRot[Y] + this->mTwr[Y], X, Z);
    float tower_ori[4];
    quat_set(tower_ori, 0, sin(0.5 * this->mTwr[Y]*0.017454), 0, cos(0.5 * this->mTwr[Y]*0.017454));

    float tower_ori_v[4];
    quat_set(tower_ori_v, sin(0.5 * this->mTwr[X]*0.017454), 0, 0, cos(0.5 * this->mTwr[X]*0.017454));
    
    quat_mul(tower_ori, tower_ori, tower_ori_v);

    float rd[2];
    rotationTo_(rd, mPos.data(), target_pos, this->mOri.data(), tower_ori);
    mPad->setAxis(cPad::MECH_TURRET_LR_AXIS, rd[X]);
    mPad->setAxis(cPad::MECH_TURRET_UD_AXIS, rd[Y]);
}

void cMech::do_fireAt() {
    //cout << "do_fireAt():\n";
    if (mPad == NULL) return;

    // Get aim-target position.
    float* target_pos = NULL;
    if (mTarget != 0) {
        cObject* target = cWorld::instance->mIndex[mTarget];
        if (target != NULL) target_pos = target->mPos.data();
        else return;
    } else return;

    // Determine nearest rotation direction
    //float rd = rotationTo(mPos, target_pos, this->mRot[Y] + this->mTwr[Y], X, Z);
    float tower_ori[4];
    quat_set(tower_ori, 0, sin(0.5 * this->mTwr[Y]*0.017454), 0, cos(0.5 * this->mTwr[Y]*0.017454));
    float rd[2];
    rotationTo_(rd, mPos.data(), target_pos, this->mOri.data(), tower_ori);
    // Fire at random and only if angle small enough.
    if (rand() % 100 <= 40 && fabs(rd[X]) < 0.5) {
        mPad->setButton(cPad::MECH_FIRE_BUTTON, true);
        //cout "fire\n";
    }
}

void cMech::do_idle() {
    if (mPad == NULL) return;
    mSpeed = 0;
    mPad->reset();
}

void cMech::do_aimFor(tSerial target) {
    mTarget = target;
}

void cMech::do_moveFor(float* dest) {
    if (dest != NULL) {
        memcpy(mDestination.data(), dest, sizeof (float) *3);
    } else {
        // Set XYZ to Not-A-Number (NaN) for no location.
        // Note that NaN-ity can only be tested either by
        // isnanf(x), !finite(x) or by x!=x as NaN always != NaN.
        mDestination[0] = nanf("");
        mDestination[1] = nanf("");
        mDestination[2] = nanf("");
        //cout << "Destination is " << ((finitef(mDestination[0])) ? "finite" : "infinite" ) << "\n";
    }
}

