/**
 * File:   cMech.h
 * Project: LinWarrior 3D
 *
 * Created on March 28, 2008, 10:23 PM (Dez1999)
 */

#ifndef _CMECH_H
#define	_CMECH_H

class cMech;

#include "cObject.h"
#include "cWeapon.h"
#include "cComcom.h"

#include <vector>

class cWeapon;

// Gamepad Mapping
#define MECH_CHASSIS_LR_AXIS  AX_LR1
#define MECH_TURRET_LR_AXIS   AX_LR2
#define MECH_TURRET_UD_AXIS   AX_UD2
#define MECH_THROTTLE_AXIS    AX_UD1
//#define MECH_FIRE_BUTTON      BT_CROSS
#define MECH_FIRE_BUTTON      BT_R2
//#define MECH_JET_BUTTON       BT_SQUARE
#define MECH_JET_BUTTON       BT_L2
#define MECH_CAMERA_BUTTON    BT_DELTA

/**
 * Models Mechlike Objects.
 * 
 */
class cMech : public cObject {
    friend class cWepcom;
    friend class cForcom;
protected:

    /// Enumeration for indexing joints in animation.

    enum Rotators {
        CTorsor, LUpLeg, RUpLeg, LLoLeg, RLoLeg, MAX_ROTATORS
    };

    /// # of camera modes: ISO C++ forbids const int with initialization >:(

    enum {
        MAX_CAMERAMODES = 6
    };

    /// Enumeration of Jumpstates for physics and animation.

    enum Jumpstates {
        GROUNDED = 0, // Sticking on ground.
        FALLING = 1, // Accellerating fall towards gravity.
        JUMPING = 2, // Gave single impulse tossing away from gravity.
        JETTING = 3 // Propelling away from gravity (fires when falling+Jet).
    };

protected:

    /// Instance counter.
    static int sInstances;

    /// Texture Binds shared between instances.
    static long sTextures[1];

protected:
    std::vector<cWeapon*> mWeapons;
    int mCurrentWeapon;

    float mTwr[3];

    // Current jumpjet set-point.
    float mJetOn;

    // Current speed set-point.
    float mSpeed;

    // Indicates current state of groundedness for physics and animation.
    int mJumpstate;

    // Joint angles for animation.
    float mRotators[MAX_ROTATORS][3];

    // Mech Joint and Mountpoint Matrices.
    float mHead[16];
    float mCenter[16];
    float mBack[16];
    float mLeft[16];
    float mRight[16];
    float mRightupper[16];
    float mRightlower[16];
    float mLeftupper[16];
    float mLeftlower[16];

    /// Current Camera mode, negative number is indicating transition.
    int mCameraState;

    cExplosion mExplosion;

    unsigned int mSoundSource;

    cComcom* mComcom;
    cTarcom* mTarcom;
    cSyscom* mSyscom;
    cWepcom* mWepcom;
    cForcom* mForcom;
    cNavcom* mNavcom;

public:
    cMech(float* pos = NULL, float* rot = NULL);
    ~cMech();

    virtual void onSpawn();

    void multEyeMatrix();
    void setAsAudioListener();

    void animatePhysics(float spf);
    void poseRunning(float spf);
    void poseJumping(float spf);

    void ChassisLR(float r);
    void ChassisUD(float r);
    float TowerLR(float r);
    void TowerUD(float r);

    void fireAllWeapons();
    void fireWeapon(unsigned n);
    void mountWeapon(char* point, cWeapon *weapon, bool add = true);

    virtual void animate(float spf);
    virtual void transform();
    virtual void drawSolid();
    virtual void drawEffect();
    virtual void drawHUD();

    virtual rImpact* damageByBeam(float* a, float* dir, float radius, float damage, bool all);
public:
    // Queries
    virtual tSerial enemyNearby();

    // Fuzzy Predicates [0,1]
    virtual float inDestinationRange();
    virtual float inMeeleRange();
    virtual float inWeaponRange();
    virtual float inTargetRange();

    // Output Actions
    virtual void do_moveTowards();
    virtual void do_moveNear();
    virtual void do_aimAt();
    virtual void do_fireAt();
    virtual void do_idle();
    virtual void do_aimFor(tSerial target);
    virtual void do_moveFor(float* dest);
};


#endif	/* _CMECH_H */

