#include "cObject.h"

#include "macros.h"

#include <vector>
using std::vector;

#include <iostream>
using std::cout;
using std::endl;


void glRotateq(float* quat) {
    /* For transposed matrix: conjugate quaternion */
    float x = -quat[X];
    float y = -quat[Y];
    float z = -quat[Z];
    float w = quat[W];
    float x2 = x*x;
    float y2 = y*y;
    float z2 = z*z;
    float m[] = {
        1 - 2 * y2 - 2 * z2, 2 * x * y - 2 * w*z, 2 * x * z + 2 * w*y, 0,
        2 * x * y + 2 * w*z, 1 - 2 * x2 - 2 * z2, 2 * y * z - 2 * w*x, 0,
        2 * x * z - 2 * w*y, 2 * y * z + 2 * w*x, 1 - 2 * x2 - 2 * y2, 0,
        0, 0, 0, 1
    };
    glMultMatrixf(m);
}


float collideLineWithBillboard(float* start3fv, float* dir3fv, float* base3fv, float radius, float height, float* impactpoint3fv, float* impactnormal3fv) {
    float near = 0.0f;
    float far = 1.0f;

    float* dir = dir3fv;

    { // Check flat circle as seen from top.

        // Calculate normalized 2d direction (forwrad) vector.
        float fwd_length = sqrt(dir[0]*dir[0] + dir[2]*dir[2]);
        if (fwd_length <= 0.0f) return -1;
        float fwd[3] = {
            dir[0] / fwd_length,
            0.0f,
            dir[2] / fwd_length
        };
        
        // Vector from 2d start to 2d billboard base center.
        float to_center[3] = {
            base3fv[0] - start3fv[0],
            0.0f,
            base3fv[2] - start3fv[2]
        };

        // Nearest distance along 2d forward vector to nearest point to center.
        float projection_scale = to_center[0]*fwd[0] + to_center[2]*fwd[2];

        // Point too far?
        float delta = projection_scale / fwd_length;
        if (delta <= near || delta >= far) return -1;

        vector_scale(impactpoint3fv, dir, delta);
        vector_add(impactpoint3fv, impactpoint3fv, start3fv);

        // Hit too high or too low?
        if (impactpoint3fv[1] < base3fv[1]) return -1;
        if (impactpoint3fv[1] >= base3fv[1]+height) return -1;

        // Nearest 2d point on 2d line to 2d billboard base circle.
        float nearest_on_line[3] = {
            fwd[0] * projection_scale,
            0.0f,
            fwd[2] * projection_scale
        };

        // Normal vector from center to nearest projected 2d point.
        float normal[3] = {
            nearest_on_line[0] - to_center[0],
            0.0f,
            nearest_on_line[2] - to_center[2],
        };

        // Missed radius?
        float normal_length_sqr = (normal[0]*normal[0] + normal[2]*normal[2]);
        if (normal_length_sqr >= radius*radius) return -1;

        vector_neg(impactnormal3fv, fwd);

        return delta;
    }
}

float collideLineWithCylinder(float* start3fv, float* dir3fv, float* base3fv, float radius, float height, float* impactpoint3fv, float* impactnormal3fv) {
    //cout << "\ncollideLineWithAABC(...)\n";

    float near = 0.0f;
    float far = 1.0f;

    float* dir = dir3fv;

    float bottom = base3fv[1];
    float top = bottom + height;

    // Missing low?
    if (start3fv[1] < bottom && dir[1] <= 0.0f) return -1;

    // Missing high?
    if (start3fv[1] > top && dir[1] >= 0.0f) return -1;

    // Find nearest entry and exit delta-distances for top and bottom pole planes.
    if (dir[1] < -0.00001f || dir[1] > +0.00001f) {
        float d1 = (bottom - start3fv[1]) / dir[1];
        float d2 = (top - start3fv[1]) / dir[1];

        bool starts_outside = (start3fv[1] < bottom) || (start3fv[1] > top);
        if (starts_outside) {
            float nearest = xmin(d1, d2);
            // Is the nearest entry point already too far away?
            if (nearest >= far) return -1;
            near = xmax(0.0f, nearest); // >0 true anyway?
        }

        float furthest = xmax(d1, d2);
        far = xmax(furthest, 1.0f);
    }

    // Check if entry point is possibly already on one of the
    // cylinder pole caps (or inside the cylinder).
    { 
        vector_set3i(impactpoint3fv, start3fv[i] + near * dir[i]);
        float xd = base3fv[0] - impactpoint3fv[0];
        float zd = base3fv[2] - impactpoint3fv[2];
        float r_sqr = xd*xd + zd*zd;
        if (r_sqr < radius*radius) {
            if (impactpoint3fv[1] < base3fv[1]) {
                vector_set(impactnormal3fv, 0.0f, -1.0f, 0.0f);
            } else {
                vector_set(impactnormal3fv, 0.0f, +1.0f, 0.0f);
            }
            return near;
        }
    }

    { // Check flat circle as seen from top.
        // Calculate normalized 2d direction (forwrad) vector.
        float fwd_length = sqrt(dir[0]*dir[0] + dir[2]*dir[2]);
        if (fwd_length <= 0.0f) return -1;
        float fwd[3] = {
            dir[0] / fwd_length,
            0.0f,
            dir[2] / fwd_length
        };

        // Vector from 2d start to 2d billboard base center.
        float to_center[3] = {
            base3fv[0] - start3fv[0],
            0.0f,
            base3fv[2] - start3fv[2]
        };

        // Nearest distance along 2d forward vector to nearest point to center.
        float projection_scale = to_center[0]*fwd[0] + to_center[2]*fwd[2];

        // Nearest 2d point on 2d line to 2d billboard base circle.
        float nearest_on_line[3] = {
            fwd[0] * projection_scale,
            0.0f,
            fwd[2] * projection_scale
        };

        // Normal vector from center to nearest projected 2d point.
        float normal[3] = {
            nearest_on_line[0] - to_center[0],
            0.0f,
            nearest_on_line[2] - to_center[2],
        };

        // Missed radius?
        float normal_length_sqr = (normal[0]*normal[0] + normal[2]*normal[2]);
        if (normal_length_sqr >= radius*radius) return -1;

        float offset = sqrt(radius * radius - normal_length_sqr);
        //cout << "offset: " << offset << "\n";

        // Point too near?
        // Use nearest from outside.
        float distance = (projection_scale - offset);
        //cout << "distance: " << distance << "\n";
        if (distance <= 0.0) return -1;

        // Point too far?
        float delta = distance / fwd_length;
        //cout << "delta: " << delta << "\n";
        if (delta <= near || delta >= far) return -1;

        vector_scale(impactpoint3fv, dir, delta);
        vector_add(impactpoint3fv, impactpoint3fv, start3fv);

        //vector_print(start3fv);
        //vector_print(impactpoint3fv);
        //vector_print(end3fv);

        impactnormal3fv[0] = impactpoint3fv[0] - base3fv[0];
        impactnormal3fv[1] = 0.0f;
        impactnormal3fv[2] = impactpoint3fv[2] - base3fv[2];
        vector_norm(impactnormal3fv, impactnormal3fv);

        return delta;
    }
}

/**
 * This function collides a line given by start3fv and end3fv with an axis aligned box.
 * The box is given through two vectors - the minumum and maximum extends: min3fv and max3fv.
 * The function returns -1 if there is no collision and returns lamda between 0.0 and 1.0 otherwise.
 * The resulting impact point is copied into impactpoint3fv.
 * If there is an impact then the following is true: impactpoint3fv = start3fv + lamda * (end3fv - start3fv).
 * Therefore the distance from start3fv to impactpoint3fv equals: |(start3fv - end3fv)| * lamda.
 * The algo here itself is somewhat tricky sorry for missing information on that one.
 */
float collideLineWithAABB(float* start3fv, float* dir3fv, float* min3fv, float* max3fv, float* impactpoint3fv, float* impactnormal3fv) {
    const float epsilon = 0.01f;

    float direction[6];
    direction[0] = direction[3] = dir3fv[0]; // Calculate Direction vector of line - and replicate it to direction[3..5], too.
    direction[1] = direction[4] = dir3fv[1]; // That replication is for easier access - no (i % 3) necessary.
    direction[2] = direction[5] = dir3fv[2];

    float dists[6];
    memcpy(&dists[0], min3fv, sizeof (float) * 3);
    memcpy(&dists[3], max3fv, sizeof (float) * 3);
    const float nrmsign[] = {-1, -1, -1, +1, +1, +1};
#if 0
    printf("Checking Planes...\n");
#endif

    loopi(6) { // For every plane of bbox.
        if (nrmsign[i] * direction[i] > 0.0) continue; // Backfacing plane?
#if 0
        printf("Checking Frontfacing %i: %.2f\n", i, direction[i]);
#endif
        if ((direction[i] == +0.0f) || (direction[i] == -0.0f)) continue; // Line is parallel to this axis.
        float lamda = (dists[i] - start3fv[i % 3]) / (direction[i]); // Calculate lamda, so that (start + (dir * lamda)) = point_on_plane.
        if (lamda < 0.0f) continue; // Point is behind line origin point.
        if (lamda > 1.0f) continue; // Point is beyond line end point.
        float p[6];
        p[0] = p[3] = start3fv[0] + direction[0] * lamda; // Calculate point on plane using previous lamda.
        p[1] = p[4] = start3fv[1] + direction[1] * lamda;
        p[2] = p[5] = start3fv[2] + direction[2] * lamda;
        int j; // Need to inspect loopvar lateron.
        for (j = 0; j < 6; j++) { // For every plane of bbox - check if point inside.
            if (j == i) {
                continue;
            } else if (j < 3) {
                if (p[j] < dists[j] - epsilon) break; // Not Inside if less than min3fv.
            } else {
                if (p[j] > dists[j] + epsilon) break; // Not Inside if greater than max3fv.
            }
        }
        if (j == 6) { // then previous test passed.
            vector_cpy(impactpoint3fv, p);
            vector_set(impactnormal3fv, 0,0,0);
            impactnormal3fv[i % 3] = nrmsign[i];
            return lamda;
        }
    }
    return -1.0f;
}

cObject::cObject() {
    //cout << "cObject::cObject()\n";

    // BASE
    mSerial = 0; // Not yet spawned, no spawndate
    mRoles = 0;
    mSeconds = 0;

    // NAMEABLE
    mName = "Anonymous";
    mDescription = "Undescribed";
    mDesignation = 0;

    // SOCIAL
    mAllies = 0;
    mEnemies = 0;

    // TRACKABLE
    mAcc.reserve(3);
    mAcc.resize(3);
    mVel.reserve(3);
    mVel.resize(3);
    mPos.reserve(3);
    mPos.resize(3);
    mRot.reserve(3);
    mRot.resize(3);
    mOri.reserve(4);
    mOri.resize(4);
    vector_set(mAcc, 0.0f, 0.0f, 0.0f);
    vector_set(mVel, 0.0f, 0.0f, 0.0f);
    vector_set(mPos, 0.0f, 0.0f, 0.0f);
    vector_set(mRot, 0.0f, 0.0f, 0.0f);
    quat_set(mOri, 0.0f, 0.0f, 0.0f, 1.0f);
    mRadius = 1;

    // DAMAGEABLE
    loopi(MAX_PARTS) {
        mHp[i] = 100.0f;
    }
    //mSplasharmor = NULL;
    //mBeamarmor = NULL;
    //mPiercearmor = NULL;

    // ENTITY
    mTarget = 0;
    mDestination.reserve(3);
    mDestination.resize(3);
    vector_set(mDestination, 0.0f, 0.0f, 0.0f);
    mPad = NULL;
    mController = NULL;
}

int cObject::ENABLE_TEXTURE_3D = 1;

cObject::cObject(cObject* source) {
    //cout << "cObject::cObject(cObject*)\n";

    // BASE
    mRoles = source->mRoles;
    mSeconds = source->mSeconds;

    // NAMEABLE
    mName = source->mName;
    mDesignation = source->mDesignation;

    // SOCIAL
    mAllies = source->mAllies;
    mEnemies = source->mEnemies;

    // TRACKABLE
    mAcc = source->mAcc;
    mVel = source->mVel;
    mPos = source->mPos;
    mRot = source->mRot;
    mRadius = source->mRadius;

    // DAMAGEABLE

    loopi(MAX_PARTS) {
        mHp[i] = source->mHp[i];
    }
    //mSplasharmor = source->mSplasharmor;
    //mBeamarmor = source->mBeamarmor;
    //mPiercearmor = source->mPiercearmor;

    // ENTITY
    mTarget = source->mTarget;
    mDestination = source->mDestination;
    //cpy3f(mDestination, source->mDestination);
    mPad = source->mPad;
    mController = source->mController;
}

float cObject::checkRelHeight() {
    float* g = cWorld::instance->mGravity.data();

    float len = 10;
    float a[] = {mPos[0], mPos[1] + 2, mPos[2]};
    float dirlen[] = {g[0] * len, g[1] * len, g[2] * len};

    float nearest2 = 1000;
    int n = 0;

    unsigned long roles = (1UL << cObject::COLLIDEABLE);
    float beamradius = 0.0f;
    float rangeradius = 25;

    std::list<cObject*>* range = cWorld::instance->filterByRange(this, a, 0.0f, rangeradius, -1);
    if (!range->empty()) {
        //cout << "range: \n" << cWorld::instance->getNames(range) << "\n";
        //cout << "range " << range->size() << "\n";
        std::list<cObject*>* role = cWorld::instance->filterByRole(this, roles, true, -1, range);
        if (!role->empty()) {
            //cout << "roles: \n" << cWorld::instance->getNames(role) << "\n";
            //cout << "roles " << role->size() << "\n";

            looplist(i, cObject*, *role) {
                cObject::rImpact* imp = ((cObject*) (*i))->damageByBeam(a, dirlen, beamradius, 0, true);
                if (imp != NULL) {
                    float v[] = {imp->pos[0] - a[0], imp->pos[1] - a[1], imp->pos[2] - a[2]};
                    float d2 = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
                    if (d2 < nearest2) nearest2 = d2;
                    n++;
                    delete imp;
                }
            }
        }
        delete role;
    }
    delete range;

    if (n > 0) {
        //cout << "below: " << n << ", " << sqrt(nearest2) << "\n";
        return sqrt(nearest2) - 2;
    }

    return mPos[Y];
}

float cObject::checkAhead(float seconds) {
    // Forward Movement
    float* dir = mVel.data();
    float velocity = sqrt(mVel[X] * mVel[X] + mVel[Y] * mVel[Y] + mVel[Z] * mVel[Z]);

    float len = 10;
    float a[] = {mPos[0], mPos[1] + 2, mPos[2]};
    float dirlen[] = {dir[0] * len, dir[1] * len, dir[2] * len};

    float nearest2 = len*len;
    int n = 0;

    unsigned long roles = (1UL << cObject::COLLIDEABLE);
    float beamradius = 0.0f;
    float rangeradius = 25;

    // Check against groundplane while descending.
    if (mVel[Y] < 0.0f) {
        //cout << " Vertical Pos and Vel: " << mPos[Y] << " | " << mVel[Y] << "\n";
        float d = a[Y] / -mVel[Y];
        float v[] = {d * mVel[X], d * mVel[Y], d * mVel[Z]};
        float d2 = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
        if (d2 < nearest2) nearest2 = d2;
    }

    std::list<cObject*>* range = cWorld::instance->filterByRange(this, a, 0.0f, rangeradius, -1, NULL);
    if (!range->empty()) {
        //cout << "range: \n" << cWorld::instance->getNames(range) << "\n";
        //cout << "range " << range->size() << "\n";
        std::list<cObject*>* role = cWorld::instance->filterByRole(this, roles, true, -1, range);
        if (!role->empty()) {
            //cout << "roles: \n" << cWorld::instance->getNames(role) << "\n";
            //cout << "roles " << role->size() << "\n";

            looplist(i, cObject*, *role) {
                // todo: min dist
                cObject::rImpact* imp = ((cObject*) (*i))->damageByBeam(a, dirlen, beamradius, 0, true);
                if (imp != NULL) {
                    float v[] = {imp->pos[0] - a[0], imp->pos[1] - a[1], imp->pos[2] - a[2]};
                    float d2 = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
                    //cout << "ahead: " << d2 << "\n";
                    if (d2 < nearest2) nearest2 = d2;
                    n++;
                    delete imp;
                }
            }
        }
        delete role;
    }
    delete range;

    float nearest = sqrt(nearest2);

    float vt = velocity * seconds;

    float d = nearest - mRadius;

    //cout << n << " ahead: " << nearest << " ahead-radius: " << d << " vel:" << v << " veltime: " << vt << "\n";
    if (vt > d && vt > 0.0f) {
        float f = d / vt;

        loopi(3) {
            mVel[i] *= f;
        }
    }

    return nearest;
}

